;   ***************************************************************
;   * Copyright (c) 2002, Embed Inc (http://www.embedinc.com)     *
;   *                                                             *
;   * Permission to copy this file is granted as long as this     *
;   * copyright notice is included in its entirety at the         *
;   * beginning of the file, whether the file is copied in whole  *
;   * or in part and regardless of whether other information is   *
;   * added to the copy.                                          *
;   *                                                             *
;   * The contents of this file may be used in any way,           *
;   * commercial or otherwise.  This file is provided "as is",    *
;   * and Embed Inc makes no claims of suitability for a          *
;   * particular purpose nor assumes any liability resulting from *
;   * its use.                                                    *
;   ***************************************************************
;
;   This module contains generic routines for handling 16F target
;   devices.  Unless otherwise noted, these routines assume the
;   target is in programming mode and ready to execute programming
;   commands.
;
         include "prg.inc"

         extern  op_reset    ;reset target, leave ready for programming
         extern  sert_send6  ;send low 6 bits of REG0 to the target
         extern  sert_send14ss ;send 14 bit REG1,REG0 data with start/stop bits
         extern  sert_recv14ss ;14 data bits from word with start/stop into REG1,REG0

         extern  done_read   ;jump here when specific read routine done
         extern  done_write  ;jump here when specific write routine done

         extern  tadr        ;requested target chip address of operation
         extern  tadri       ;actual internal target chip address
         extern  cntwait     ;wait interval down counter
         extern  tprog       ;wait time after write in timer 2 ticks
         extern_flags        ;declare global flag bits EXTERN
;
;***********************************************************************
;
;   Configuration constants.
;
lbank    equ     0           ;register bank for the local state of this module
;
;   Derived constants.
;
lbankadr equ     bankadr(lbank) ;address within local state register bank
;
;***********************************************************************
;
;   Global state.  All this state is assumed to be in the GBANK register
;   bank by other modules.
;
.bank#v(gbank) udata

;
;***********************************************************************
;
;   Local state.
;
  if lbank != gbank
.bank#v(lbank) udata
    endif

.16f     code
;
;***********************************************************************
;
;   Local subroutine SET_ADR
;
;   Set the internal target chip address to match the desired address.
;   The desired address is in TADR and the current address internal
;   to the target chip is TADRI.  The PIC16 chips only implement a 14
;   bit address space, so the routines in this module only use the
;   low 2 bytes of TADR and TADRI.
;
         locsub  set_adr, regf0 | regf1
;
;   Perform a reset if the current target address is unknown.  A reset
;   unconditionally sets the target address to 0.
;
         dbankif gbankadr
         btfss   flag_tadrkn ;the target address is known ?
         jump    sa_reset    ;no, reset the target chip to force known address
;
;   Perform a reset if the current address is greater than the desired
;   address.  This resets the current address to 0, and is the only way
;   of decreasing the program counter.
;
         dbankif gbankadr
         movf    tadri+1, w  ;get current address high byte
         subwf   tadr+1, w   ;compare to desired address high byte
         skip_wle            ;not definitely past desired address ?
         jump    sa_reset    ;definitely past desired address, do the reset
         skip_z              ;high byte compare inconclusive ?
         jump    sa_dreset   ;definitely before desired address, skip the reset

         movf    tadri+0, w  ;get current address low byte
         subwf   tadr+0, w   ;compare to desired address low byte
         skip_wgt            ;past desired address, need to do reset ?
         jump    sa_dreset   ;no, skip the reset
;
;   The current address is greater than the desired address.  Reset
;   the target chip, which also resets the current address to zero.
;
sa_reset
         gcall   op_reset    ;reset the target, curr address becomes 0
         wait                ;wait for the reset to complete
         dbankif gbankadr

sa_dreset dbankis gbankadr   ;skip here to avoid doing a reset
;
;   The current target chip address is less than or equal to the desired
;   address.
;
;   Now check for the current address is in the low half of memory and the
;   desired address in the upper half.  The program counter can't just be
;   incremented from 1FFFh to 2000h.  The LOAD CONFIGURATION command is
;   required to switch the program counter to 2000h.
;
         dbankif gbankadr
         btfss   tadr+1, 5   ;desired address is in high half of adr space ?
         jump    sa_dhhalf   ;no, skip switching to high half
         btfsc   tadri+1, 5  ;current address is in low half of adr space ?
         jump    sa_dhhalf   ;no, skip switching to high half

         dbankif gbankadr
         loadk16 tadri, h'2000' ;update our copy of internal chip address
         clrf    reg0        ;pass LOAD CONFIGURATION opcode
         gcall   sert_send6  ;send the command
         clrf    reg1        ;pass data word to send after the command
         gcall   sert_send14ss ;send REG1,REG0 as data word

sa_dhhalf unbank             ;skip to here to avoid switching to high addresses
;
;   The current address is less than or equal to the desired address,
;   and is in the same half of the address space.
;
;   Now increment the current address until it matches the desired address.
;
         dbankif gbankadr
sa_loopinc                   ;back here after each address increment
         movf    tadri+0, w  ;get current address low byte
         xorwf   tadr+0, w   ;compare to desired address low byte
         skip_z              ;low bytes match, need to check high bytes ?
         jump    sa_doinc    ;addresses don't match, do an increment
         movf    tadri+1, w  ;get current address high byte
         xorwf   tadr+1, w   ;compare to desired address high byte
         skip_nz             ;addresses don't match ?
         jump    sa_doneinc  ;current address = desired address, all done
         ;
         ;   The current and desired addresses don't match.  Increment
         ;   the current address.
         ;
sa_doinc
         movlw   b'000110'   ;pass INCREMENT ADDRESS opcode
         movwf   reg0
         gcall   sert_send6  ;send it

         dbankif gbankadr
         incf    tadri+0     ;increment low byte of current address
         skip_nz             ;no carry into upper byte ?
         incf    tadri+1     ;propagate the carry
         jump    sa_loopinc  ;back to compare addresses after the increment

sa_doneinc dbankis gbankadr  ;done incrementing current address to match desired

         leaverest
;
;***********************************************************************
;
;   Routine F16_READ
;
;   Read the target chip data at address TADR.  The data is returned in the
;   low bits of REG1,REG0 with unused high bits set to 0.
;
;   REG2, REG3 may be trashed.
;
         glbent  f16_read

         mcall   set_adr     ;make sure target chip is at the desired address

         dbankif gbankadr
         movlw   b'000100'   ;get READ DATA FROM PROGRAM MEMORY command
         btfsc   flag_spdata ;really are in program memory space ?
         movlw   b'000101'   ;get READ DATA FROM DATA MEMORY command
         movwf   reg0
         gcall   sert_send6  ;send it

         gcall   sert_recv14ss ;get the 14 data bits into low bits of REG1,REG0

         gjump   done_read
;
;***********************************************************************
;
;   Routine WRITE_12F6
;
;   Write the data in REG1,REG0 to the target chip at address TADR.
;
;   This version is for the target chips 12F629, 12F675.
;
;   WARNING:  This routine jumps into routine F16_WRITE, which then
;     jumps to DONE_WRITE when done.
;
;   REG2, REG3 may be trashed.
;
         glbent  write_12f6

         dbankif gbankadr
         bsf     flag_op1    ;indicate to use 12F6xx commands
         jump    f16_write_comm ;to common part of subroutine F16_WRITE
;
;***********************************************************************
;
;   Routine F16_WRITE
;
;   Write the data in REG1,REG0 to the target chip at address TADR.
;
;   This version is for generic 16Fxxx targets.
;
;   REG2, REG3 may be trashed.
;
         glbent  f16_write

         dbankif gbankadr
         bcf     flag_op1    ;indicate to use generic 16F commands
;
;**********
;
;   Common code for performing a write on a 14 bit core PIC.  Some state
;   has been set to indicate details of the PIC variant:
;
;     FLAG_OP1  -  Set if this is 12Fxxx chip with 14 bit core (like 12F629),
;       reset for generic 14 bit core (like 16F877).
;
f16_write_comm unbank
         movf    reg0, w     ;save original REG0 value in REG2
         movwf   reg2

         mcall   set_adr     ;make sure target chip is at the desired address

         dbankif gbankadr
         movlw   b'000010'   ;get LOAD DATA FOR PROGRAM MEMORY command
         btfsc   flag_spdata ;really are in program memory space ?
         movlw   b'000011'   ;get LOAD DATA FOR DATA MEMORY command
         movwf   reg0
         gcall   sert_send6  ;send the command
         movf    reg2, w     ;restore the original data in REG0
         movwf   reg0
         gcall   sert_send14ss ;send the data word

         dbankif gbankadr
         movlw   b'011000'   ;get 16F87x command BEGIN PROGRAM ONLY CYCLE
         btfsc   flag_op1    ;really is for 16F87x?
         movlw   b'001000'   ;get 12F6xx command BEGIN PROGRAMMING
         movwf   reg0
         gcall   sert_send6

         dbankif gbankadr
         movf    tprog, w    ;get timer 2 ticks to wait for program cycle done
         intr_off            ;temp disable interrupts
         movwf   cntwait+0   ;set the wait time
         clrf    cntwait+1
         bcf     flag_done   ;indicate a wait interval is in progress
         intr_on             ;re-enable interrupts

         writing             ;indicate the target chip is being written to

         movf    reg2, w     ;restore the original data in REG0
         movwf   reg0

         gjump   done_write

         end
